/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */

package au.com.terabit.jproactor;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.UnknownHostException;
import java.nio.channels.ServerSocketChannel;
//import java.nio.channels.SocketChannel;

/**
 * The <code>Acceptor</code> class allows to start user defined server-side
 * protocol servers. 
 * <br>Acceptor opens {@link java.net.ServerSocketChannel} and requests
 * {@link Demultiplexor} to execute accept operation, once accept is completed
 * registered {@link AsynchChannelHandlerFactory} instantiates a new {@link Protocol}
 * object and starts execution using the same multiplexor instance.
 * <p>
 * The following code fragment shows how to start test EchoServer using acceptor:
 * <blockquote><pre>
 * <i>// Create and run Demultiplexor instance for EchoServer on port 9999 </i> 
 * Demultiplexor m = new Demultiplexor();
 * m.start();
 *
 * <i>// Create acceptor instance for EchoServer on port 9999 </i> 
 * Acceptor a = new Acceptor( m, new EchoServerProtocolFactory(), 9999 );
 * <i>// start listening for incoming connections</i>
 * a.start();
 * </pre></blockquote>
 *
 * @author Alexander Libman
 * @author Yevgeny Libman
 * @version <code>$Revision$ $Date$</code>
 */
public class Acceptor implements AsynchAcceptHandler, AsynchChannelHandler
{

    /** How may accepts to perform, default value */
    public final static int NUM_INITIAL_ACCEPTS = 5;
    
    /** Acceptors' socket address */
    SocketAddress m_address = null;

    /** Server socket channel for this acceptor */
    ServerSocketChannel m_serverChannel = null;
    
    /**
     * Procotol factory assigned for this socket address 
     */
    AsynchChannelHandlerFactory m_factory = null;

    /**
     * Demultiplexor instance 
     */
    Demultiplexor m_multiplexor = null;
    
    /**
     * IO m_asynchChannel instance
     */
    TCPAsynchChannel  m_achannel = null;
    

    // ------------------------------------------------------------------------
    // Public methods

    /**
     * Creates acceptor for the specified <code>AsynchHandlerFactory</code>
     * and on the specified tcp/ip port.
     * 
     * @param m - multiplexor instance
     * @param factory - protocol factory
     * @param port - tcp/ip port to listen on
     */
    public Acceptor( Demultiplexor m, 
                     int port,
                     AsynchChannelHandlerFactory factory)
        throws UnknownHostException
    {
        this( m, 
              new InetSocketAddress( InetAddress.getLocalHost(), port ),
              factory);
    }

    /**
     * Creates acceptor for the specified <code>AsynchHandlerFactory</code>
     * and on the specified tcp/ip port and host.
     * 
     * @param m - multiplexor instance
     * @param factory - protocol factory
     * @param port - tcp/ip port to listen on
     * @param host - host name
     */
    public Acceptor( Demultiplexor m, 
                     int port, 
                     String host,
                     AsynchChannelHandlerFactory factory)
    {
        this( m, new InetSocketAddress( host, port ), factory );
    }
    
    /**
     * Creates acceptor for the specified <code>AsynchHandlerFactory</code>
     * and on the specified <code>Socket Address</code>.
     * 
     * @param m - multiplexor instance
     * @param factory - protocol factory
     * @param address - socket address to listen on
     */
    public Acceptor( Demultiplexor m, 
                     SocketAddress address,
                     AsynchChannelHandlerFactory factory)
    {
        this.m_factory = factory;
        this.m_address = address;
        this.m_multiplexor = m;
        
        try
        {
            // create channel for accepting connections
            // and bind it to the specified address
            this.m_serverChannel = ServerSocketChannel.open();
            this.m_serverChannel.socket().setPerformancePreferences (1,0,0);
            this.m_serverChannel.socket().setReuseAddress(true);
            this.m_serverChannel.socket().bind( address );
        
            // Create TCPAsynchChannel for accepting connections
            this.m_achannel = new TCPAsynchChannel (m_multiplexor, m_serverChannel);
            
            this.m_achannel.setChannelHandler(this);
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    
    //
    // Interface AsynchAcceptHandler implementation
    //
    
 
    /**
     * Implementation of {@link AsynchChannelHandler#acceptCompleted(OpAccept op)}.
     * Executes accept operation, non-blocking, instantiates protocol on the
     * accepted channel, starts protocol using acceptor's multiplexor instance.
     * 
     * @throws Exception if this method is called
     */
    public void acceptCompleted( OpAccept op ) throws Exception
    {
    	
        AsynchChannel asynchChannel = op.getAcceptedChannel();
        
        if (asynchChannel != null)
        {
            // start accept operation again
        	m_achannel.accept( this );
        	
            // now handle new connection
            AsynchChannelHandler protocol = m_factory.createChannelHandler ();

            asynchChannel.setChannelHandler (protocol);
        }
    }

    //
    // Interface AsynchChannelHandler implementation
    //
    
    public void channelAttached( AsynchChannel channel ) throws Exception
    {
    }
    
    public void channelClosed( AsynchChannel channel ) throws Exception
    {
        System.out.println( "Acceptor Closed");
    }

    /**
     * Starts acceptor, using {@link #NUM_INITIAL_ACCEPTS} default number of
     * accept operations.
     * <br>
     * This method can be called more than once.
     *  
     * @see au.com.terabit.multiplexor.AsynchChannelHandler#start()
     * @see #start(int numInitialAccepts)
     */
    public void start() throws Exception
    {
        start (NUM_INITIAL_ACCEPTS);
    }
    
    /**
     * Starts accept operation.
     * <br>
     * This method can be called more than once.
     *
     * @throws IOException if error occurs, the operation won't be started
     *                      in this case.
     */
    public void start(int numInitialAccepts) throws Exception
    {
        for ( int i = 0; i < numInitialAccepts; ++i )
        {
            m_achannel.accept( this  );
        } 
    }

}