/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.jproactor;

import java.net.SocketAddress;
import java.nio.ByteBuffer;

import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.atomic.*;


/**
 * The <code>AsynchChannel</code> is an intermidiate interface that serves as
 * mediator between <code>Demultiplexor</code> and <code>AsynchHandler</code>
 * <p>
 * All implemented protocols should be based on this interface.
 *
 * @author <a href="mailto:libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 * 
 * @see IOOperation
 */
public abstract class AsynchChannel 
{

    /** Channel states */
    public enum State { OPENED, CONNECTED, CLOSING, CLOSED }


    /* ---------------------------------------------------------------------- */
    //
    // instance variables
    
    /** Internal lock for synchronization */
    protected ReentrantLock m_lock = new ReentrantLock ();
    
    /** This channel's handler, protocol instance */ 
    private  AtomicReference<AsynchChannelHandler> m_channelHandler = 
        new AtomicReference<AsynchChannelHandler> (null);
    
    /** channelState : OPENED, CONNECTED, CLOSING, CLOSED */
    protected State m_channelState = State.OPENED;
 
    /** queue for the requested read operations */
    protected ConcurrentLinkedQueue<IOOperation> m_readQue = 
        new ConcurrentLinkedQueue<IOOperation>();

    /** queue for the requested write operations */
    protected ConcurrentLinkedQueue<IOOperation> m_writeQue = 
        new ConcurrentLinkedQueue<IOOperation>();
    
    /** queue for the completed write operations */
    //protected ConcurrentLinkedQueue<IOOperation> m_completedQue = 
    //    new ConcurrentLinkedQueue<IOOperation>();
    
    public IOStatistic m_readInfo  = new IOStatistic ();
    public IOStatistic m_writeInfo = new IOStatistic ();


    /* ---------------------------------------------------------------------- */
    
    /**
     * Creates <code>AsynchChannel</code> instance given specified 
     * <code>Demultiplexor</code>, <code>AsynchHandler</code> and 
     * <code>SelectableChannel</code> instances.
     *
     * @param m <code>Demultiplexor</code> instance this m_asynchChannel works with
     * @param p <code>AsynchHandler</code> protocol instance
     * @param sc <code>SelectableChannel</code> instance this m_asynchChannel operates on
     */
    public AsynchChannel()
    {
    }
    
    public AsynchChannelHandler setChannelHandler (AsynchChannelHandler handler)
    {
        AsynchChannelHandler h = m_channelHandler.getAndSet(handler);
        try
        {
            handler.channelAttached (this);
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        return h;
    }
    
   /**
    * Returns AsynchCloseHandler instance for this AsynchChannel
    * 
    * @return m_closeHandler
    */
    public AsynchChannelHandler getChannelHandler ()
    {
        return m_channelHandler.get();
    }


    protected abstract void startRead   (OpRead    op) throws Exception;
    protected abstract void startWrite  (OpWrite   op) throws Exception;
    protected abstract void startAccept (OpAccept  op) throws Exception;
    protected abstract void startConnect(OpConnect op)  throws Exception;
    public    abstract void startTimer  (OpTimer   op)  throws Exception;
      
    
    protected abstract boolean startClose  ();

    /**
     * this method is should be called by derived implemnations
     * with locked m_lock 
     * 
     * @return true  if channel is just or already closed
     *         false otherwise
     */
    protected boolean checkForClose ()
    {
    	AsynchChannelHandler handler = null;
        switch (m_channelState)
        {
        case CLOSING:
             m_channelState = State.CLOSED;
             handler = m_channelHandler.getAndSet(null);
             break;
        case CLOSED:
             return true;
		default:
			 return false;
        }
        
        if (handler != null)
        {
        	m_lock.unlock();
        	try
        	{
        		handler.channelClosed(this);
        	}
        	catch (Exception e)
        	{
        		e.printStackTrace();
        	}
        	finally 
        	{
        		m_lock.lock();
        	}
        }
        return true;
    }
    
    /**
     * 
     * @param op
     * @return > 0 number bytes read
     *         0   not finished yet
     *         < 0 errors or end of data
     * @throws Exception
     */
    protected abstract int finishRead(OpRead op) throws Exception;
    /**
     * 
     * @param op
     * @return > 0 number bytes written
     *         0   not finished yet
     *         < 0 errors or peer closed
     * @throws Exception
     */
    protected abstract int finishWrite(OpWrite op) throws Exception;
    
    /**
     * 
     * @param op
     * @return new accepted AsynchChannel
     *         null not finished yet
     * @throws Exception
     */
    protected abstract AsynchChannel finishAccept  (OpAccept op) throws Exception;
    
    /**
     * 
     * @param op
     * @return true if connect finished
     *         false if not
     *         may be to fix: 
     *         1  finished OK,
     *         0  not finished yet 
     *         -1 finished with errors
     * @throws Exception
     */
    protected abstract boolean finishConnect (OpConnect op) throws Exception;
    

    /**
     * @param completedQue
     */
    protected boolean dispatchCompletions (ConcurrentLinkedQueue<IOOperation> completedQue)
    {
        boolean rc = true;
        for (;;)
        {
            // take operation from the queue 
            IOOperation op = completedQue.poll();
            if (op == null)
            {
                break;
            }
            
            try
            {
                op.onComplete();
            }
            catch (Exception e)
            {
                 e.printStackTrace();
                 rc = false;
            }
        }
        return rc;
    }
 
    /**
     * 
     * @param reqQue
     * @return null if all operations in reqQue are executed
     *         first operation that is still in reqQue and was not executed
     */
    protected IOOperation executeListAndDispatch (ConcurrentLinkedQueue<IOOperation> reqQue) 
    {
        IOOperation op = null ;
        boolean   flgCancel = isClosed ();    
        
        for (; (op = reqQue.peek())!= null; )
        {
            if (!executeOp (op, flgCancel))
            {
                break;
            }
            
            reqQue.poll();  // remove it
            m_lock.unlock();
            try
            {
                op.onComplete();
            }
            catch (Exception e)
            {
                 e.printStackTrace();
                 flgCancel = true;
            }
            finally
            {
                m_lock.lock();
            }
        }
        
        return op;
    }

    /**
     * 
     * @param reqQue
     * @param completedQue
     * @return null if all operations in reqQue are executed
     *         first operation that is still in reqQue and was not executed
     */
    protected IOOperation executeListAndAdd (ConcurrentLinkedQueue<IOOperation> reqQue,
                                             boolean flgCancel,
                                             ConcurrentLinkedQueue<IOOperation> completedQue) 
    {
        IOOperation op = null ;
        
        for (; (op = reqQue.peek())!= null;)
        {
            if (!executeOpAndAdd (op, flgCancel, completedQue))
            {
                break;
            }
            reqQue.poll();  // remove it
        }
        
        return op;
    }
    
    protected boolean executeOpAndAdd (IOOperation op,
                                       boolean flgCancel,
                                       ConcurrentLinkedQueue<IOOperation> completedQue)
    {
        if (executeOp (op, false))
        {
            completedQue.add(op);
            return true;
        }
        return false;
    }
    
    
    protected boolean executeOp (IOOperation op, boolean flgCancel)
    {
        
        switch (op.m_state)
        {
        case IOOperation.STATE_FINISHED:
        case IOOperation.STATE_CANCELED:
            return true;
        }
            
        if (flgCancel || isClosed ())
        {
             op.setCancelled(null);
             return true;
        }
        
        return op.execute();
    }

   
    protected void enqueueOperation(IOOperation op, 
                                    ConcurrentLinkedQueue<IOOperation> que,
                                    IOStatistic info) throws Exception
    {
        if (isClosed ())
        {
            throw new Exception (
                "AsynchChannel::startOperation: attempt to start IO on inactive channel");
        }
        
        op.m_state = IOOperation.STATE_QUEUED;
        que.add(op);
        ++info.m_opStarted;

    }
    
    
    /**
     * Request for read operation.
     * This is main interface to start asynchronous read operation.
     * 
     * @param buf buffer to read data to
     * @return OpRead operation - future result
     * @throws exception if error occurs
     */ 
    public OpRead read( ByteBuffer buf, AsynchReadHandler handler ) throws Exception
    {
        OpRead op = new OpRead( this, handler, buf, false );
        
        m_lock.lock();
        try
        {
            enqueueOperation (op, m_readQue, m_readInfo);
            startRead (op);
        }
        finally 
        {
            m_lock.unlock();
        }
        
        return op;
    }
    
    /**
     * Request to execute non-partitional read.
     * Read exactly {@link ByteBuffer#limit()} bytes to the buffer.
     * 
     * @param buf 
     * @return OpRead operation - future result
     * @throws exception if error occurs
     */
    public OpRead read_n ( ByteBuffer buf, AsynchReadHandler handler  ) throws Exception
    {
        OpRead op = new OpRead( this, handler, buf, true );
 
        m_lock.lock();
        try
        {
            enqueueOperation (op, m_readQue, m_readInfo);
            startRead (op);
        }
        finally 
        {
            m_lock.unlock();
        }
        
        return op;
    }

    /**
     * Requets for write operation.
     * This is main interface to start asynchronous write operation.
     * 
     * @param buf buffer to write data from
     * @return OpWrite operation - future result
     * @throws exception if error occurs
     */ 
    public OpWrite write( ByteBuffer buf, AsynchWriteHandler handler  ) throws Exception
    {
        OpWrite op = new OpWrite( this, handler, buf, false );

        m_lock.lock();
        try
        {
            enqueueOperation (op, m_writeQue, m_writeInfo);
            startWrite (op);
        }
        finally 
        {
            m_lock.unlock();
        }
        
        return op;
    }
    
    /**
     * Request to execute non-partitional write.
     * Write exactly {@link ByteBuffer#limit()} bytes to the associated channel.
     * 
     * @param buf buffer to write data from
     * @return OpWrite operation - future result
     * @throws exception if error occurs
     */
    public OpWrite write_n( ByteBuffer buf, AsynchWriteHandler handler  ) throws Exception
    {
        OpWrite op = new OpWrite( this, handler, buf, true );

        m_lock.lock();
        try
        {
            enqueueOperation (op, m_writeQue, m_writeInfo);
            startWrite (op);
        }
        finally 
        {
            m_lock.unlock();
        }
        
        return op;
    }

    /**
     * Request to execute accept operation.
     * 
     * @return OpAccept operation - future result
     * @throws Exception if error occurs
     */
    public OpAccept accept( AsynchAcceptHandler handler) throws Exception
    {
        OpAccept op = new OpAccept( this, handler );
        
        m_lock.lock();
        try
        {
            enqueueOperation (op, m_readQue, m_readInfo);
            startAccept (op );
        }
        finally 
        {
            m_lock.unlock();
        }
        
        return op;
    }

    /**
     * Request to execute connect operation.
     * 
     * @param SocketAddress  address to connect
     * @return OpConnect operation - future result
     * @throws Exception if error occurs
     */
    public OpConnect connect (SocketAddress addr, AsynchConnectHandler handler) throws Exception
    {
        OpConnect op = new OpConnect (this, handler, addr);
                
        m_lock.lock();
        try
        {
            enqueueOperation (op, m_writeQue, m_writeInfo);
            startConnect (op);
        }
        finally 
        {
            m_lock.unlock();
        }
        
        return op;
    }
    
    
    /**
     * Schedules timeout action to be called back on the associated 
     * <code>AsynchHandler</code> instance.
     * <p>
     * Note: while IO operations will be delivered to the handler in a
     * serialized way, timer notifications might be delivered concurrently.    
     * 
     * @param delay delay in milliseconds before timeout action is to be executed. 
     * @return OpTimer operation - future result
     */
    public OpTimer schedulerTimer(long delay, AsynchTimerHandler handler) throws Exception
    {
        if ( delay < 0 )
            throw new IllegalArgumentException( "Negative timeout delay" );

        OpTimer op = new OpTimer(delay , handler);
        
        startTimer (op);
 
        return op;
    }
    
    public void close() 
    {
        m_lock.lock();
        try
        {
            if (setClosing ())
            {
                startClose ();
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        finally 
        {
            m_lock.unlock();
        }
        
    }
    
 
    public boolean isOpened()    { return m_channelState == State.OPENED ||
                                          m_channelState == State.CONNECTED; }
    public boolean isConnected() { return m_channelState == State.CONNECTED; }
    
    public boolean isClosed ()  { return m_channelState == State.CLOSING ||
                                         m_channelState == State.CLOSED; }
    public boolean isCloseStarted()    { return m_channelState == State.CLOSING; }
    public boolean isCloseFinished ()   { return m_channelState == State.CLOSED; }

    // -----------------------------------------------------------------------
    //
    // Package visible methods
    

    protected final boolean setConnected ()
    {
        switch  (m_channelState)
        {
        case OPENED:
            m_channelState = State.CONNECTED;
            return true;
        }
        return false;
    }
   
    
    protected final boolean setClosing ()
    {
        switch  (m_channelState)
        {
        case OPENED:
        case CONNECTED:
            m_channelState = State.CLOSING;
            return true;
        }
        return false;
    }

}
