/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
 
package au.com.terabit.jproactor;
import java.io.IOException;
//import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.spi.SelectorProvider;
import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.ReentrantLock;

/**
 * The <code>Demultiplexor</code> class is the core engine of the <i>JavaProactor</i>.
 * <p>
 * Demultiplexor acts as IO processor and  manages all IO operations requested
 * by {@link TCPAsynchChannel} class. The IO oprations are proccessed asynchrously by
 * the pool of threads using leader/follower pattern. Number of processing
 * thread can be configured by {@link Demultiplexor#start(int)} operation. 
 * <p>
 * Usually client applications does not require direct access to mutiplexor
 * class, only when multiple protocols would like to reuse the same multiplexor
 * instance.
 * <p>
 * Each multiplexor instance instantiates one {@link java.nio.channels.Selector}
 * instance.
 *
 * @author Alexander Libman
 * @author Yevgeny Libman
 * @version <code>$Revision$ $Date$</code>
 */
public class Demultiplexor implements Runnable
{
    
    // ----------------------------------------------------------------------
    // Static section
    
    /** Default Thread Pool size */
    public static final int POOL_SIZE = 5;
    
    // ----------------------------------------------------------------------
    // Instance variables
    
    public static final int STRATEGY_FAST = 0;
    public static final int STRATEGY_BALANCE = 1;
    
    static public int dispatchStategy_ = STRATEGY_FAST;
    
    
    /** Selector - should work in separate thread or in the user thread */
    private Selector selector_ = null;
    
    /** Iterator for set of ready keys */
	Iterator<SelectionKey>  itrKey_ = null;
   
    /** Iterator of the set of expired timers from the timer Queue */
    Iterator<OpTimer>  itrTimer_ = null;

    /** Stop flag */
    private AtomicBoolean stopFlag_ = new AtomicBoolean (false);
    
    /** Requests Queue - collection of IOHandlers */
    private ConcurrentLinkedQueue<TCPAsynchChannel> requestQue_ = 
    	new ConcurrentLinkedQueue<TCPAsynchChannel> ();
    
  
    /** total number threads  */
    private int    numThreads1_ = 0;
    /** number of threads runnning event loop */
    private int    numThreads2_ = 0;
    private Object lockMultiplexor_ = new Object();
    
    /** number threads doing user upcalls */
    private AtomicInteger numDispatchers_ = new AtomicInteger (0);
     
    /** number threads waiting for job */
    private AtomicInteger numFollowers_ = new AtomicInteger (0);
    
    /** number handlers ready for dispatch job */
    private int numReadyHandlers_ = 0;
    
    /** leader thread */
	private AtomicReference<DemultiplexorThread> leader_ = 
		new AtomicReference<DemultiplexorThread> (null);
	
	/** lock to obtain leadership - to avoid conditional variable */
	private	ReentrantLock leaderLock_ = new ReentrantLock ();
	
	private TimerQueue timerQue_ = new TimerQueue();


    private class DemultiplexorThread implements Runnable
    {
    	
        public static final int ST_ZERO     = 0;
        public static final int ST_FOLLOWER = 1;
        public static final int ST_LEADER   = 2; 
        public static final int ST_HANDLER  = 4;
        public static final int ST_TIMER    = 8;
        public static final int ST_STOP     = 16;
        
        
        
        int       m_state    = ST_ZERO;
        TCPAsynchChannel m_readyChannel  = null;
        int       m_readyEvents = 0;
        int       m_remainEvents = 0;
        OpTimer   m_timer = null;

        // ----------------------------------------------------------------------
        // Runnable Interface

        public void run()
        {
        	for (;m_state != ST_STOP; )
            {       
        		m_state = ST_FOLLOWER;
        		numFollowers_.incrementAndGet();
                		
        		leaderLock_.lock();
        		try
        		{
        			// Leader -only one thread at moment 
        			// can execute this block
        			
        			//if (leader.compareAndSet(null, this))
        			leader_.set(this);
        			m_state = ST_LEADER;
        			numFollowers_.decrementAndGet();
        			m_state = getEventForDispatch ();  
        		}
        		finally
        		{
        			leader_.set (null);
        			leaderLock_.unlock();
        		}
        		
        		if (m_state == ST_HANDLER)
        		{
                	numDispatchers_.incrementAndGet();
        			m_readyChannel.dispatchEvents (m_readyEvents, m_remainEvents);
        			numDispatchers_.decrementAndGet();
        			m_readyChannel = null;
        		}
                else if (m_state == ST_TIMER)
                {
                    m_timer.onComplete();
                    m_timer = null;
                }
            }
        }
        
        private int getEventForDispatch ()
        {
			while (!stopFlag_.get())
			{
                // if there are no current operations to execute then wait on select
			    if (itrKey_ == null && itrTimer_ == null)
	    		{
			    	if (!updateInterestSet ())
			    	{
			    		// m_readyChannel contains cancelled m_asynchChannel
			    		return ST_HANDLER;
			    	}
			    	
			    	waitOnSelect ();
	    			numReadyHandlers_ = selector_.selectedKeys().size();
	    			itrKey_ = selector_.selectedKeys().iterator();
                    itrTimer_ = timerQue_.getExpiredTimers().iterator();
	    		}
	          		                	
	    		if (itrKey_ != null && itrKey_.hasNext ())
	    		{
	    			SelectionKey key = itrKey_.next();
	                
	    			// retrieve associated TCPAsynchChannel instance
	    			// and ready events mask
	    			m_readyChannel = (TCPAsynchChannel) key.attachment();
	    			m_readyEvents = key.readyOps();
                    // remove ready events from interest mask. 
	    			m_remainEvents = key.interestOps() & ~m_readyEvents;
	    			
	    			// this call won't block as there is no active select at this point
	    			key.interestOps (m_remainEvents);
	                
	    			itrKey_.remove ();
	    			--numReadyHandlers_;
	    			return ST_HANDLER;
	    		}
                itrKey_ = null;

                // check expired timers iterator
                if (itrTimer_ != null && itrTimer_.hasNext())
                {
                    m_timer = itrTimer_.next();
                    itrTimer_.remove();
                    return ST_TIMER;
                }
                itrTimer_ = null;
			}
			return ST_STOP;
		}         
            
        private boolean updateInterestSet()
        {
        	TCPAsynchChannel achannel = null;
                  	
            try
            {
            	while ((achannel = requestQue_.poll()) != null)
            	{
            		if (!achannel.updateInterest (selector_))
            		{
            			m_readyChannel = achannel;
			    		m_remainEvents = 0;
			    		m_readyEvents = IOOperation.OP_ALL;
            			return false;
            		}
            	}
        	
            }
            catch ( Exception e ) 
            { 
                e.printStackTrace(); 
            }
            return true;
        }
        
         
        
        private int waitOnSelect () 
        {
            int rc = 0;
            long delay = timerQue_.getFirstDelay();
                  	
            if ( delay < 0 )
            {
                System.out.println( "Internal error: waitOnSelect() - negative delay" );
                // return immediately to process expired timers
                return rc;
            }
            try
            {
                rc = selector_.select( delay );
                 
                if (rc == 0)  
                {
                   //System.out.println ("Select returned 0");
                }
            }
            catch ( Exception e ) 
            { 
                e.printStackTrace(); 
            }
            return rc;
        }
        
		
    } // DemultiplexorThread 
		
    // ----------------------------------------------------------------------
    // Public methods

    /**
     * Creates multipliexor instance with default pool size.
     * 
     * @throws IOException
     */
    public Demultiplexor() throws IOException 
    {
        selector_ = SelectorProvider.provider().openSelector();
    }
    
    /**
     * Start multiplexor.
     * 
     * @param poolSize thread pool size 
     */
    public void start (int poolSize)
    {
    	synchronized (lockMultiplexor_)
    	{
    		for (int i=0; i < poolSize && ! stopFlag_.get(); ++i)
    		{
    			new Thread (this).start();
    			++numThreads1_;
    		}
    	}
    }

    /**
     * Start multiplexor using default thread pool size.
     *
     * @see Demultiplexor#POOL_SIZE
     */
    public void start ()
    {
        start( POOL_SIZE );
    }


    // ----------------------------------------------------------------------
    // Runnable Interface

    public void run()
    {
    	
        synchronized (lockMultiplexor_)
        {
        	++numThreads2_;
    	
        	System.out.println ( "Enter Demultiplexor::run thread="
                                 + Thread.currentThread().getName()
                                 + " total=" + numThreads2_);                                
        }
        
    	DemultiplexorThread thread = new DemultiplexorThread ();
    	
    	thread.run();
    	
        
        synchronized (lockMultiplexor_)
        {
        	--numThreads1_;
        	--numThreads2_;
            System.out.println ( "Leave Demultiplexor::run thread="
                    + Thread.currentThread().getName()
                    + " total=" + numThreads2_);      

        	lockMultiplexor_.notifyAll();
        }
    }
    
    /**
     * Register m_asynchChannel.
     */
    public void declareInterest(TCPAsynchChannel achannel) 
    {
    	requestQue_.add (achannel);
    	if (leader_.get() != null)
    	{
    		selector_.wakeup ();
    	}
    }
    
    /**
     * Demultiplexor shutdown. Stops all IO operations.
     * 
     */
    public void shutdown() throws InterruptedException 
    {
        stopFlag_.set(true);
        selector_.wakeup ();
        
        // wait for all threads
    	synchronized (lockMultiplexor_)
        {
            while (numThreads1_ > 0)
            {
                try
                {
                    lockMultiplexor_.wait();
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }
        }
    	
        try
        {
        	TCPAsynchChannel achannel;
        	
        	itrKey_ = selector_.keys().iterator();
        	while (itrKey_.hasNext ())
    		{
    			SelectionKey key = (SelectionKey) itrKey_.next();
                
    			// retrieve associated TCPAsynchChannel instance
    			// and ready events mask
    			achannel = (TCPAsynchChannel) key.attachment();
    	 		achannel.onSelectorClosed (selector_);
                achannel.dispatchEvents(IOOperation.OP_ALL, 0);
    		}
            
            while ((achannel = requestQue_.poll()) != null)
            {
                achannel.onSelectorClosed (selector_);
                achannel.dispatchEvents(IOOperation.OP_ALL, 0);
            }
   	
        }
        catch ( Exception e ) 
        { 
            e.printStackTrace(); 
        }
        
        try
        {
            // close the Selector
            selector_.close();
        }
        catch (Exception e)
        {
            e.printStackTrace(); 
        }
    }
    
    /**
     * Schedules timer operation.
     * 
     * @param timer operation
     */
    public void startTimer( final OpTimer timer )
    {
        if ( timerQue_.add( timer ) )
        {
            if ( leader_.get() != null )
            {
                selector_.wakeup ();
            }
        }
    }

}
